from .models import model_exists, CHAT_MODELS

VERBOSITY_LEVELS = [0, 1, 2]

class Config():
    def __init__(self, chat_model: str, verbosity: int = 1, explain_graph: bool = True,
            save_graph_as_file: bool = False):
        assert model_exists(chat_model), f"Model name must be one of: {CHAT_MODELS}"
        self.chat_model = chat_model
        assert verbosity in VERBOSITY_LEVELS, f"Verbosity must be one of: {VERBOSITY_LEVELS}"
        self.verbosity = verbosity
        assert type(explain_graph) is bool, f"Explain graph flag must be boolean type."
        self.explain_graph = explain_graph
        assert type(save_graph_as_file) is bool, f"Save graph as file flag must be boolean type."
        self.save_graph_as_file = save_graph_as_file

        if verbosity >= 1:
            print(f"Verbosity set to {verbosity}.")



import openai
import time

def get_answer_from_model(user_prompt: str, model: str, 
        messages: list[dict[str, str]]) -> str:
    messages.append({"role": "user", "content": user_prompt})

    # TODO: Pass number of attempts as parameters
    # TODO: Pass retry time as parameter
    max_attempts = 5
    for i in range(1, max_attempts+1):
        try:
            print("Getting answer from model...")
            response = openai.ChatCompletion.create(
                model=model, messages=messages,
            )

        except openai.error.RateLimitError as e:
            retry_time = e.retry_after if hasattr(e, 'retry_after') else 5
            print(f"Rate limit exceeded. Retrying in {retry_time} seconds...")
            time.sleep(retry_time)

        except openai.error.APIError as e:
            retry_time = e.retry_after if hasattr(e, 'retry_after') else 5
            print(f"API error occurred. Retrying in {retry_time} seconds...")
            time.sleep(retry_time)

        except OSError as e:
            retry_time = 5
            print(f"Connection error occurred: {e}. Retrying in {retry_time} seconds...")      
            time.sleep(retry_time)
        
        else:
            break

    answer = response["choices"][0]["message"]["content"]
    return answer



import json
from typing import Any, Callable

from ..recognizers import AbstractRecognizer
from .. import regex
from ..config import Config

# next commands field indexes
NEXT_COMMAND_ID = 0
DEPENDENT_ON_DATA = 1
REQUIRED_VALUE = 2

class CommandNode:
    def __init__(self, data: dict, commands: dict[str, dict], 
            command_name_to_func: dict[str, Callable]):
        self.id = data["id"]
        self.command_name = data["name"]

        assert self.command_name in commands, f"Command '{self.command_name}' does not exist."
        assert self.command_name in command_name_to_func, f"Command '{self.command_name}' does not have a function declaration."

        self.command = command_name_to_func[self.command_name]

        self.arguments = data["arguments"]
        self.next_commands: list[list[int | str | Any]] = data["next_commands"]

        self.data_generated = None

    def __str__(self):
        return f"CommandNode(id={self.id}, command={self.command})"

    def execute_command(self, config: Config, graph, arguments: dict[str, Any]):
        print(f"\n\nRunning '{self.command_name}' command with id {self.id}...")
        if config.verbosity >= 2:
            print(f"Using arguments: {arguments}")
        self.data_generated = self.command(config, graph, **arguments)
    
    def get_next_commands_to_execute(self) -> list[int]:
        next_commands_to_execute = []
        for next_command in self.next_commands:
            next_command_id = next_command[NEXT_COMMAND_ID]
            dependent_on_data = next_command[DEPENDENT_ON_DATA]
            required_value = next_command[REQUIRED_VALUE]

            if dependent_on_data is None: # doesn't matter the result of current node
                next_commands_to_execute.append(next_command_id)
            else: # next command execution depends on the result of current node
                if self.data_generated[dependent_on_data] == required_value:
                    next_commands_to_execute.append(next_command_id)
        return next_commands_to_execute

class Graph:
    def __init__(self, recognizer: AbstractRecognizer, commands_data_str: str):
        self.set_start_data(recognizer, commands_data_str)
        self.initialize()

    def set_start_data(self, recognizer: AbstractRecognizer, commands_data_str: str):
        self.recognizer = recognizer
        self.commands_data_str = commands_data_str
        self.commands = recognizer.commands
        self.command_name_to_func = recognizer.command_name_to_func

    def build_graph(self, commands_data_str: str):
        self.set_start_data(self.recognizer, commands_data_str)

        graph_build_data = generate_graph_build_data(commands_data_str)
        (self.commands_data_str_by_node,
         self.commands_data,
         self.data_references_in_each_command) = graph_build_data

        self.build_nodes()

    def build_node(self, command_data: dict) -> CommandNode:
        node = CommandNode(command_data, self.commands, self.command_name_to_func)
        if node.id in self.reached_nodes_ids:
            # recover results of the node
            reached_node = self.nodes[node.id]
            node.data_generated = reached_node.data_generated
        return node

    def build_nodes(self):
        for command_data in self.commands_data.values():
            node = self.build_node(command_data)
            self.nodes[node.id] = node

    def initialize(self):
        self.reached_nodes_ids: list[int] = []
        self.nodes: dict[str, CommandNode] = {}
        self.build_graph(self.commands_data_str)
            
    def execute_node(self, node_id: int, config: Config) -> list[int]:
        node = self.nodes[node_id]

        # inject data references to current node's arguments
        injected_command_data_str = self.commands_data_str_by_node[node.id]
        for node_to_inject_id in self.data_references_in_each_command[node.id]:
            node_to_inject = self.nodes[node_to_inject_id]
            # update node data
            injected_command_data_str = regex.inject_node_data(
                injected_command_data_str, node_to_inject.id,
                node_to_inject.data_generated,
            )
        injected_command_data = get_node_command_data(injected_command_data_str)

        # update node
        node = self.build_node(injected_command_data)
        self.nodes[node.id] = node

        node.execute_command(config, self, node.arguments)
        self.reached_nodes_ids.append(node.id)
        
        next_commands_to_execute = node.get_next_commands_to_execute()
        return next_commands_to_execute

    def print_graph(self, explain_graph: bool):
        print("\n\n--- Commands graph ---")

        print("\n~~ Graph ~~")
        for node in self.nodes.values():
            print(f"\n{node.id}. {node.command_name}")
            
            if node.next_commands:
                print(f"\tCommands executed by this node:")
                for next_command_id, dependent_on_data, required_value in node.next_commands:
                    next_node = self.nodes[next_command_id]
                    print(f"\t{next_node.id}. {next_node.command_name}")
                    if dependent_on_data is not None:
                        print(f"\t\tIf '{dependent_on_data}' generated data has value: {required_value}.")

        if explain_graph:
            print("\n~~ Explanation ~~")
            explanation = self.recognizer.explain_graph_in_natural_language(self.commands_data_str)
            print(explanation)

        print("\n--- -------------- ---\n")

    def execute_commands(self, config: Config):
        self.initialize()
        if config.verbosity >= 1:
            self.print_graph(config.explain_graph)

        first_node_id = sorted(self.nodes.keys())[0]
        next_commands_to_execute = self.execute_node(first_node_id, config)
        new_next_commands_to_execute = []
        while True:
            new_next_commands_to_execute.clear()
            for next_command_id in next_commands_to_execute:
                new_next_commands_to_execute.extend(self.execute_node(next_command_id, config))
            next_commands_to_execute = new_next_commands_to_execute.copy()

            if not next_commands_to_execute:
                break

def get_node_data_references(command_data_str: str) -> dict:
    data_references = regex.find_data_references_indices(command_data_str)
    return data_references

def unescape_node_arguments(arguments: dict):
    for name, value in arguments.items():
        if type(value) is str:
            arguments[name] = regex.unescape_str_in_json(value)
        # TODO: Unescape strings inside of lists/dicts/tuples arguments.

def get_node_command_data(command_data_str: str) -> dict:
    nullified_data_str = regex.nullify_all_data_references(command_data_str)
    command_data_as_list = json.loads(nullified_data_str, strict=False)

    id_ = command_data_as_list[0]
    name = command_data_as_list[1]
    arguments = command_data_as_list[2]

    unescape_node_arguments(arguments)

    next_commands: list[list[int | str | Any]] = command_data_as_list[3]
    assert all(
        map(
            lambda next_command:
                (type(next_command[0]) is int and 
                    type(next_command[1]) in [str, type(None)]),
            next_commands,
        )
    ), f"Next commands field must match data types: list[list[int, str, Any]].\n{next_commands}"

    command_data = {
        "id": id_,
        "name": name,
        "arguments": arguments,
        "next_commands": next_commands,
    }

    return command_data

def generate_graph_build_data(commands_data_str: str):
    """
    Parses a commands data string to a JSON to create the graph data

    Args:
        commands_data_str (str): A string containing the data to create a graph/command.

    Returns:
        a tuple: Containing:
            commands_data_str_by_node (dict of str): Commands data as a string.
                Keys are the ID of the commands; values are the commands data
                as strings.
        
            commands_data: data of the commands as JSON (to create a graph).

            data_references_in_each_command: data references that appear inside of each command (data that will be injected).
                A key is the ID of a command. The value stores a dictionary of data references inside of that particular command.
    """
    data_references_in_each_command = {}

    commands_data = {}
    commands_data_str_by_node = {}
    for line_num, command_data_str in enumerate(commands_data_str.splitlines(), start=1):
        data_references = get_node_data_references(command_data_str)
        try:
            command_data = get_node_command_data(command_data_str)
        except Exception as e:
            print(f"!!! Can't decode command data string to JSON in line {line_num}: {command_data_str}")
            raise e

        command_id = command_data["id"]

        commands_data_str_by_node[command_id] = command_data_str
        commands_data[command_id] = command_data
        data_references_in_each_command[command_id] = data_references

    return commands_data_str_by_node, commands_data, data_references_in_each_command


from typing import Any, Callable

from ..chat import get_answer_from_model
from ..config import Config
from .graphs import Graph
from ..util.math_expr import safe_eval_math_expr

ESSENTIAL_COMMANDS = {
    "THINK": {
        "description": "Generates a thought in your 'mind' without writing it down yet. This command allows you to create new ideas, reflect, and analyze information. For example, when asked to 'Write a three-paragraph article about Lenz's Law,' you would first use this command to generate the article and then use another to write it.",
        "arguments": {
            "about": {"description": "What to think about. Example: 'Three-paragraph article about Lenz's Law.'", "type": "string"},
        },
        "generates_data": {
            "thought": {"description": "Text generated by thinking.", "type": "string"},
        },
    },
    "IF": {
        "description": "Returns the Boolean value of a condition.",
        "arguments": {
            "condition": {"description": "Condition. Can be in natural language.", "type": "string"},
        },
        "generates_data": {
            "result": {"description": "Result of the condition: 0 or 1.", "type": "boolean"},
        },
    },
    "CALCULATE": {
        "description": "Evaluates a mathematical expression in a str. Supports +, -, *, /, %, **, //. Can be entered in natural language.",
        "arguments": {
            "expression": {"description": "Math expression. '(-1) ** (1/2)', 'Negative one raised to 1/2.'", "type": "string"},
        },
        "generates_data": {
            "result": {"description": "Result of evaluation", "type": "int or float or complex"},
        },
    },
    # TODO: Create a FOR command to increment a counter variable
}

# Commands functions
# Must be named 'LowercaseCommandName_command'
# The first argument must be the Config object, followed by the Graph object
# The arguments must match the arguments from the ESSENTIAL_COMMANDS dictionary
# The return value must be a dictionary which keys must match the "generates_data" keys
# The data types must match the ones declared in the ESSENTIAL_COMMANDS dictionary

def think_command(config: Config, graph: Graph, about: str) -> dict[str, Any]:
    messages = [
        {
            "role": "system", 
            "content": "You are a model used when executing a 'THINK' command, which function is to reflect, think, write, or ideate. Only do what the prompt says; DO NOT add useless/extra information/irrelevant chat/irrelevant explanation."
        },
    ]
    thought = get_answer_from_model(about, config.chat_model, messages)

    results = {
        "thought": thought,
    }
    return results

def if_command(config: Config, graph: Graph, condition: str) -> dict[str, Any]:
    messages = [
        {
            "role": "system", 
            "content": f"You are a model that evaluates conditions, both in natural language and symbolic language. Given a condition, you respond with the number «1» (true) or «0» (false). DO NOT write ANYTHING ELSE EVER.",
        },
    ]
    result = get_answer_from_model(condition, config.chat_model, messages)
    try:
        result = bool(result)
    except Exception as e:
        print(f"Could not convert result from IF command '{result}' to boolean.")
        raise e

    results = {
        "result": result,
    }
    return results

def calculate_command(config: Config, graph: Graph, expression: str) -> dict[str, Any]:
    try:
        result = safe_eval_math_expr(expression)
    except ValueError:
        messages = [
            {
                "role": "system",
                "content": f"You are a model that takes a math expression in natural language and returns ONLY the math expression, without any words. You can only use +, -, *, /, %, **, //. Example: 'Square root of negative one plus eight' -> '(-1) ** (1/2) + 8'"
            }
        ]
        expression_ = get_answer_from_model(expression, config.chat_model, messages)
        result = safe_eval_math_expr(expression_)

    results = {
        "result": result,
    }
    return results

ESSENTIAL_COMMAND_NAME_TO_COMMAND_FUNC = {
    key: eval(f"{key.lower()}_command")
    for key in ESSENTIAL_COMMANDS
}

def get_command(command_name: str) -> Callable:
    """
    Returns the function corresponding to the name.

    Args:
        command_name (str): Name of the function.

    Returns:
        function: The corresponding function.
    """    
    return ESSENTIAL_COMMAND_NAME_TO_COMMAND_FUNC[command_name]

def add_commands(commands: dict[str, dict], command_name_to_func: dict[str, dict],
        new_commands: dict[str, dict], new_command_name_to_func: dict[str, Callable]):
    commands.update(new_commands)
    command_name_to_func.update(new_command_name_to_func)

def add_essential_commands(commands: dict[str, dict], command_name_to_func: dict[str, dict]):
    add_commands(commands, command_name_to_func, ESSENTIAL_COMMANDS, ESSENTIAL_COMMAND_NAME_TO_COMMAND_FUNC)



from typing import Callable
from .config import Config

from .chat import get_answer_from_model

class AbstractRecognizer():
    def __init__(self, config: Config, commands: dict[str, dict], 
            command_name_to_func: dict[str, Callable], 
            recognition_messages: list[dict[str, str]], 
            explanation_messages: list[dict[str, str]]):
        """
        Args:
            config: Config object.
            commands: The commands that the model can use, 
                described in natural language.
            command_name_to_func: Dictionary that takes
                the name of a command and returns the corresponding function.
            recognition_messages: List of dictionaries
                that contain system messages explaining the LLM the format
                of its answers and the commands it can use while recognizing.
            explanation_messages: List of dictionaries
                that contain system messages explaining the LLM the format
                of its answers and the commands it can use while explaining
                in natural language a graph.
        """
        self.config = config
        self.commands = commands
        self.command_name_to_func = command_name_to_func
        self.recognition_messages = recognition_messages
        self.explanation_messages = explanation_messages

    def recognize(self, instruction: str) -> str:
        """
        Analyzes an instruction and creates data to create a graph of commands
        that will fulfill the instruction.
        """
        print(f"Input tokens used by messages (instruction recognition): ~{len(str(self.recognition_messages)) / 4} tokens.")

        commands_data_str = get_answer_from_model(instruction, self.config.chat_model, self.recognition_messages)

        if self.config.verbosity >= 2:
            print(f"\n\n~ ~ ~ ~ Commands data generated by the LLM\n\n{commands_data_str}\n\n~ ~ ~ ~")
        
        if self.config.save_graph_as_file:
            with open("graph.txt", "w+") as f:
                f.write(commands_data_str)
                f.close()

        return commands_data_str

    def explain_graph_in_natural_language(self, commands_data_str: str) -> str:
        """
        Takes a graph as string and uses natural language to explain the
        connections in the graph.
        """
        print(f"Input tokens used by messages (graph explanation): ~{len(str(self.explanation_messages)) / 4} tokens.")

        explanation = get_answer_from_model(commands_data_str, self.config.chat_model, self.explanation_messages)

        return explanation

class ComplexRecognizer(AbstractRecognizer):
    def __init__(self, config: Config, commands: dict[str, dict], 
            command_name_to_func: dict[str, Callable]):
        """
        Recognizer that creates a complex graph of commands, using conditions,
        data references from other commands and executes multiple commands.
        Each command can execute multiple commands based on multiple conditions.
        """        
        recognition_messages = [
            {
            'role': 'system', 
            'content':
                """You are a tool that, based on the user's prompt, detects the series of commands that must be executed, arguments that each command will have, and the relationship between each command (for example, what data generated by a command will be used as an argument for another)."""
                """\n*IMPORTANT*: *WRITE in the LANGUAGE that the USER writes his/her prompt in*."""
                """\n*IMPORTANT*: You can ONLY use the given commands. **NEVER try to use OTHERS**."""
                """\n*IMPORTANT*: While creating the graph, you are NOT talking to the user. You are JUST CREATING THE GRAPH, so do not write suggestions for a user inside of the graph. Create the complete graph by your own."""

                f"""\n\nCommands:\n{commands}"""

                """\n\nDepending on the prompt, you will use different commands and different arguments and relationships between commands. The only way you can see data generated by other commands from a command is by passing them as arguments."""

                """\n\n*Your response will have this format, ALWAYS stick to it*:"""
                """\n[command_id, "COMMAND_NAME", {"arg1": value1, "arg2": value2, ...}, [[next_command_id, "dependent_on_data", required_value], [...], ...]]"""
                """\n[...]"""
                """\n\nThe structure is JSON-Lines-like (so, use double quotes to create str, etc.)"""
                """\nONLY use newline characters to separate the data of each node (as it's a JSON-Lines-like format)."""
                """\nYou will replace EVERYTHING in the format with the correct data."""
                """\nTo reference data generated by a command, use __&i.data__. 'i' is the ID of the command; 'data' is the name of the generated data."""
                """\nAlways consider the DATA TYPE of the data you are referencing (e.g. string-types can not have indices like '__&i.thought[j]__'; data does not have attributes unless specified)."""
                """\nYou can ONLY reference data in the arguments of the nodes."""
                """\nThe structure will ultimately be like a graph. The next_command_id defines which command will be executed next; dependent_on_data is the name of the data generated by the current command which will be used as a condition (*null if the next command will be executed no matter what the results of the current command where*), and required_value is what value the dependent_on_data must have to execute the next command (null if dependent_on_data was also null)."""
                """\nNote that dependent_on_data is NOT a data reference for the next command; it's just what will be used as a CONDITION to determine if the next command will be executed."""
                """\ndependent_on_data field CAN NOT be 'dependent_on_data'. It must be null if it doesn't matter."""
                """\nA command can execute multiple next commands."""
                # """\nBe concise but *solid* with the structure."""
                """\n*Consider that the user might write incorrectly and their inputs might be ambiguous*."""
                """\nThe only way to reference the data of other commands is by using the __&i.data__ referencing. Commands DO NOT KNOW each other's data."""
                """\nProvide all the relevant context in the arguments of the commands, so that you're not ambiguous."""
                """\nNote that each command DOES NOT know about the other commands. You have to use DATA REFERENCES as ARGUMENTS to pass context."""
                """\nBe creative and logic when using the commands' arguments and data references."""

                """\n\nFor example:"""

                """\n\nExample 1:"""
                """\nUser prompt: 'Write an article about Lenz's Law, copy it to my clipboard and save it as a file.'"""
                """\nYour response might be: '[1, "THINK", {"about": "Article about Lenz's Law"}, [[2, null, null]]]\n[2, "WRITE_CLIPBOARD", {"content": "__&1.thought__"}, [[3, null, null]]]\n[3, "WRITE_FILE", {"content": "__&1.thought__", "file_name": "Article about Lenz's Law"}, []]'"""
                """\n\nExample 2:"""
                """\nUser prompt: 'Read my clipboard, write a scientific article about the content in it, and save it as a file with a name related to the article.'"""
                """\nYour response might be: '[1, "READ_CLIPBOARD", {}, [[2, null, null]]]\n[2, "THINK", {"about": "Scientific article about __&1.content__"}, [[3, null, null]]]\n[3, "THINK", {"about": "Name including extension to save a file about this article: __&2.thought__"}, [[4, null, null]]]\n[4, "WRITE_FILE", {"content": "__&2.thought__", "file_name": "__&3.thought__"}, []]'"""
                """\n\nExample 3:"""
                """\nUser prompt: 'Look up the best courses on ASP.Net on Google and show me the first page that appears, if it is relevant.'"""
                """\nYour response might be: '[1, "SEARCH_GOOGLE", {"query": "best courses on ASP.Net"}, [[2, null, null]]]\n[2, "READ_WEBPAGE", {"url": "__&1.urls[0]__"}, [[3, null, null]]]\n[3, "IF", {"condition": "Is this a relevant course on ASP.Net? __&2.text__"}, [[4, "result", 1], [5, "result", 0]]]\n[4, "WRITE_TO_USER", {"content": "Relevant course: __&1.urls[0]__"}, []]\n[5, "WRITE_TO_USER", {"content": "It is not relevant."}, []]'"""
            }
        ]
        explanation_messages = [
            {
            'role': 'system', 
            'content':
                """You are a tool that, given a graph of commands, explains in natural language what the graph does, how the nodes connect, and all the details about the graph, commands and nodes."""
                """\n*IMPORTANT*: *WRITE in the LANGUAGE that the USER writes his/her prompt in*."""

                f"""\n\nCommands:\n{commands}"""

                """\n\nThe graph of commands has this format:"""
                """\n[command_id, "COMMAND_NAME", {"arg1": value1, "arg2": value2, ...}, [[next_command_id, "dependent_on_data", required_value], [...], ...]]"""
                """\n[...]"""
                """\n\nThe structure is JSON-Lines-like."""
                """\nTo reference data generated by a command, use __&i.data__. 'i' is the ID of the command; 'data' is the name of the generated data."""
                """\nThe structure will ultimately be like a graph. The next_command_id defines which command will be executed next, dependent_on_data is the name of the data generated by the current command which will be used as a condition (null if it won't be used), and required_value is what value the dependent_on_data must have to execute the next command (or null if it doesn't matter)."""
                """\nA command can execute multiple next commands."""

                """\n\nExamples of graphs of commands generated from the user prompt:"""

                """\n\nExample 1:"""
                """\nUser prompt: 'Write an article about Lenz's Law, copy it to my clipboard and save it as a file.'"""
                """\nThe response might be: '[1, "THINK", {"about": "Article about Lenz's Law"}, [[2, null, null]]]\n[2, "WRITE_CLIPBOARD", {"content": "__&1.thought__"}, [[3, null, null]]]\n[3, "WRITE_FILE", {"content": "__&1.thought__", "file_name": "Article about Lenz's Law"}, []]'"""
                """\n\nExample 2:"""
                """\nUser prompt: 'Read my clipboard, write a scientific article about the content in it, and save it as a file with a name related to the article.'"""
                """\nThe response might be: '[1, "READ_CLIPBOARD", {}, [[2, null, null]]]\n[2, "THINK", {"about": "Scientific article about __&1.content__"}, [[3, null, null]]]\n[3, "THINK", {"about": "Name including extension to save a file about this article: __&2.thought__"}, [[4, null, null]]]\n[4, "WRITE_FILE", {"content": "__&2.thought__", "file_name": "__&3.thought__"}, []]'"""
                # """\n\nExample 3:"""
                # """\nUser prompt: 'Look up the best courses on ASP.Net on Google and show me the first page that appears, if it is relevant.'"""
                # """\nThe response might be: '[1, "SEARCH_GOOGLE", {"query": "best courses on ASP.Net"}, [[2, null, null]]]\n[2, "READ_WEBPAGE", {"url": "__&1.urls[0]__"}, [[3, null, null]]]\n[3, "IF", {"condition": "Is this a relevant course on ASP.Net? __&2.text__"}, [[4, "result", 1], [5, "result", 0]]]\n[4, "WRITE_TO_USER", {"content": "Relevant course: __&1.urls[0]__"}, []]\n[5, "WRITE_TO_USER", {"content": "It is not relevant."}, []]'"""
            }
        ]
        super().__init__(config, commands, command_name_to_func, recognition_messages, explanation_messages)

class SequentialRecognizer(AbstractRecognizer):
    def __init__(self, config: Config, commands: dict[str, dict], 
            command_name_to_func: dict[str, Callable]):
        """
        Recognizer that creates a sequential graph of commands (without conditions),
        using data references from other commands and executes multiple commands.
        Each command can only execute the next command, and there are no conditions
        to execute the next commands.
        """        
        recognition_messages = [
            {
            'role': 'system', 
            'content':
                """You are a tool that, based on the user's prompt, detects the series of commands that must be executed, arguments that each command will have, and the relationship between each command (for example, what data generated by a command will be used as an argument for another)."""
                """\n*IMPORTANT*: *WRITE in the LANGUAGE that the USER writes his/her prompt in*."""
                """\n*IMPORTANT*: You can ONLY use the given commands. **NEVER try to use OTHERS**."""
                """\n*IMPORTANT*: While creating the graph, you are NOT talking to the user. You are JUST CREATING THE GRAPH, so do not write suggestions for a user inside of the graph. Create the complete graph by your own."""

                f"""\n\nCommands:\n{commands}"""

                """\n\nDepending on the prompt, you will use different commands and different arguments and relationships between commands. The only way you can see data generated by other commands from a command is by passing them as arguments."""

                """\n\n*Your response will have this format, ALWAYS stick to it*:"""
                """\n[command_id, "COMMAND_NAME", {"arg1": value1, "arg2": value2, ...}, [[next_command_id, null, null]]]"""
                """\n[...]"""
                """\n\nThe structure is JSON-Lines-like (so, use double quotes to create str, etc.)"""
                """\nONLY use newline characters to separate the data of each node (as it's a JSON-Lines-like format)."""
                """\nYou will replace EVERYTHING in the format with the correct data."""
                """\nTo reference data generated by a command, use __&i.data__. 'i' is the ID of the command; 'data' is the name of the generated data."""
                """\nAlways consider the DATA TYPE of the data you are referencing."""
                """\nYou can ONLY reference data in the arguments of the nodes."""
                """\nThe structure is a sequential list of commands. The next_command_id defines which command will be executed next."""
                """\nBe concise but *solid* with the structure."""
                """\n*Consider that the user might write incorrectly and their inputs might be ambiguous*."""
                """\nThe only way to reference the data of other commands is by using the __&i.data__ referencing. Commands DO NOT KNOW each other's data."""
                """\nProvide all the relevant context in the arguments of the commands, so that you're not ambiguous."""
                """\nNote that each command DOES NOT know about the other commands. You have to use DATA REFERENCES as ARGUMENTS to pass context."""
                """\nBe creative and logic when using the commands' arguments and data references."""

                """\n\nFor example:"""

                """\n\nExample 1:"""
                """\nUser prompt: 'Write an article about Lenz's Law, copy it to my clipboard and save it as a file.'"""
                """\nYour response might be: '[1, "THINK", {"about": "Article about Lenz's Law"}, [[2, null, null]]]\n[2, "WRITE_CLIPBOARD", {"content": "__&1.thought__"}, [[3, null, null]]]\n[3, "WRITE_FILE", {"content": "__&1.thought__", "file_name": "Article about Lenz's Law"}, []]'"""
                """\n\nExample 2:"""
                """\nUser prompt: 'Read my clipboard, write a scientific article about the content in it, and save it as a file with a name related to the article.'"""
                """\nYour response might be: '[1, "READ_CLIPBOARD", {}, [[2, null, null]]]\n[2, "THINK", {"about": "Scientific article about __&1.content__"}, [[3, null, null]]]\n[3, "THINK", {"about": "Name including extension to save a file about this article: __&2.thought__"}, [[4, null, null]]]\n[4, "WRITE_FILE", {"content": "__&2.thought__", "file_name": "__&3.thought__"}, []]'"""
                """\n\nExample 3:"""
                """\nUser prompt: 'I'll tell you a math operation and you'll tell me the result.'"""
                """\nYour response might be: '[1, "REQUEST_USER_INPUT", {"message": "Please enter the math operation you want me to calculate."}, [[2, null, null]]]\n[2, "CALCULATE", {"expression": "__&1.input__"}, [[3, null, null]]]\n[3, "WRITE_TO_USER", {"content": "The result of the math operation is: __&2.result__"}, []]'"""
            }
        ]
        explanation_messages = [
            {
            'role': 'system', 
            'content':
                """You are a tool that, given a graph of commands, explains in natural language what the graph does, how the nodes connect, and all the details about the graph, commands and nodes."""
                """\n*IMPORTANT*: *WRITE in the LANGUAGE that the USER writes his/her prompt in*."""

                f"""\n\nCommands:\n{commands}"""

                """\n\nThe graph of commands has this format:"""
                """\n[command_id, "COMMAND_NAME", {"arg1": value1, "arg2": value2, ...}, [[next_command_id, null, null]]]"""
                """\n[...]"""
                """\n\nThe structure is JSON-Lines-like."""
                """\nTo reference data generated by a command, use __&i.data__. 'i' is the ID of the command; 'data' is the name of the generated data."""
                """\nThe structure is a sequential list of commands. The next_command_id defines which command will be executed next."""

                """\n\nExamples of graphs of commands generated from the user prompt:"""

                """\n\nExample 1:"""
                """\nUser prompt: 'Write an article about Lenz's Law, copy it to my clipboard and save it as a file.'"""
                """\nThe response might be: '[1, "THINK", {"about": "Article about Lenz's Law"}, [[2, null, null]]]\n[2, "WRITE_CLIPBOARD", {"content": "__&1.thought__"}, [[3, null, null]]]\n[3, "WRITE_FILE", {"content": "__&1.thought__", "file_name": "Article about Lenz's Law"}, []]'"""
                """\n\nExample 2:"""
                """\nUser prompt: 'Read my clipboard, write a scientific article about the content in it, and save it as a file with a name related to the article.'"""
                """\nThe response might be: '[1, "READ_CLIPBOARD", {}, [[2, null, null]]]\n[2, "THINK", {"about": "Scientific article about __&1.content__"}, [[3, null, null]]]\n[3, "THINK", {"about": "Name including extension to save a file about this article: __&2.thought__"}, [[4, null, null]]]\n[4, "WRITE_FILE", {"content": "__&2.thought__", "file_name": "__&3.thought__"}, []]'"""
                """\n\nExample 3:"""
                """\nUser prompt: 'I'll tell you a math operation and you'll tell me the result.'"""
                """\nThe response might be: '[1, "REQUEST_USER_INPUT", {"message": "Please enter the math operation you want me to calculate."}, [[2, null, null]]]\n[2, "CALCULATE", {"expression": "__&1.input__"}, [[3, null, null]]]\n[3, "WRITE_TO_USER", {"content": "The result of the math operation is: __&2.result__"}, []]'"""

            }
        ]
        super().__init__(config, commands, command_name_to_func, recognition_messages, explanation_messages)

class SingleRecognizer(AbstractRecognizer):
    def __init__(self, config: Config, commands: dict[str, dict], 
            command_name_to_func: dict[str, Callable]):
        """
        Recognizes only one command in an instruction.
        """
        recognition_messages = [
            {
            'role': 'system', 
            'content':
                """You are a tool that, based on the user's prompt, detects which command must be executed."""
                """\n*IMPORTANT*: *WRITE in the LANGUAGE that the USER writes his/her prompt in*."""
                """\n*IMPORTANT*: You can ONLY use the given commands. **NEVER try to use OTHERS**."""
                """\n*IMPORTANT*: You can ONLY use ONE command. You can't use multiple."""
                """\n*IMPORTANT*: While writing, you are JUST WRITING THE COMMAND'S DATA, so do not write suggestions for the user."""

                f"""\n\nCommands:\n{commands}"""

                """\n\n*Your response will have this format, ALWAYS stick to it*:"""
                """\n[command_id, "COMMAND_NAME", {"arg1": value1, "arg2": value2, ...}, []]"""
                """\n\nThe structure is JSON-Lines-like (so, use double quotes to create str, etc.)"""
                """\nYou will replace EVERYTHING in the format with the correct data."""
                """\n*Consider that the user might write incorrectly and their inputs might be ambiguous*."""
                """\nProvide all the relevant context in the arguments of the command, so that you're not ambiguous."""
                """\nBe creative and logic when using the commands' arguments."""

                """\n\nExample (using multiple commands; you can only use ONE command):"""

                """\n\nExample 1:"""
                """\nUser prompt: 'Write an article about Lenz's Law, copy it to my clipboard and save it as a file.'"""
                """\nYour response might be: '[1, "THINK", {"about": "Article about Lenz's Law"}, [[2, null, null]]]\n[2, "WRITE_CLIPBOARD", {"content": "__&1.thought__"}, [[3, null, null]]]\n[3, "WRITE_FILE", {"content": "__&1.thought__", "file_name": "Article about Lenz's Law"}, []]'"""
                """\n\nExample 2:"""
                """\nUser prompt: 'Read my clipboard, write a scientific article about the content in it, and save it as a file with a name related to the article.'"""
                """\nYour response might be: '[1, "READ_CLIPBOARD", {}, [[2, null, null]]]\n[2, "THINK", {"about": "Scientific article about __&1.content__"}, [[3, null, null]]]\n[3, "THINK", {"about": "Name including extension to save a file about this article: __&2.thought__"}, [[4, null, null]]]\n[4, "WRITE_FILE", {"content": "__&2.thought__", "file_name": "__&3.thought__"}, []]'"""
                """\n\nExample 3:"""
                """\nUser prompt: 'I'll tell you a math operation and you'll tell me the result.'"""
                """\nYour response might be: '[1, "REQUEST_USER_INPUT", {"message": "Please enter the math operation you want me to calculate."}, [[2, null, null]]]\n[2, "CALCULATE", {"expression": "__&1.input__"}, [[3, null, null]]]\n[3, "WRITE_TO_USER", {"content": "The result of the math operation is: __&2.result__"}, []]'"""
            }
        ]
        explanation_messages = [
            {
            'role': 'system', 
            'content':
                """You are a tool that, given a graph of commands, explains in natural language what the graph does, how the nodes connect, and all the details about the graph, commands and nodes."""
                """\n*IMPORTANT*: *WRITE in the LANGUAGE that the USER writes his/her prompt in*."""

                f"""\n\nCommands:\n{commands}"""

                """\n\nThe graph of commands has this format:"""
                """\n[command_id, "COMMAND_NAME", {"arg1": value1, "arg2": value2, ...}, [[next_command_id, null, null]]]"""
                """\n[...]"""
                """\n\nThe structure is JSON-Lines-like."""
                """\nTo reference data generated by a command, use __&i.data__. 'i' is the ID of the command; 'data' is the name of the generated data."""
                """\nThe structure is a sequential list of commands. The next_command_id defines which command will be executed next."""

                """\n\nExamples of graphs of commands generated from the user prompt:"""

                """\n\nExample 1:"""
                """\nUser prompt: 'Write an article about Lenz's Law, copy it to my clipboard and save it as a file.'"""
                """\nThe response might be: '[1, "THINK", {"about": "Article about Lenz's Law"}, [[2, null, null]]]\n[2, "WRITE_CLIPBOARD", {"content": "__&1.thought__"}, [[3, null, null]]]\n[3, "WRITE_FILE", {"content": "__&1.thought__", "file_name": "Article about Lenz's Law"}, []]'"""
                """\n\nExample 2:"""
                """\nUser prompt: 'Read my clipboard, write a scientific article about the content in it, and save it as a file with a name related to the article.'"""
                """\nThe response might be: '[1, "READ_CLIPBOARD", {}, [[2, null, null]]]\n[2, "THINK", {"about": "Scientific article about __&1.content__"}, [[3, null, null]]]\n[3, "THINK", {"about": "Name including extension to save a file about this article: __&2.thought__"}, [[4, null, null]]]\n[4, "WRITE_FILE", {"content": "__&2.thought__", "file_name": "__&3.thought__"}, []]'"""
                """\n\nExample 3:"""
                """\nUser prompt: 'I'll tell you a math operation and you'll tell me the result.'"""
                """\nThe response might be: '[1, "REQUEST_USER_INPUT", {"message": "Please enter the math operation you want me to calculate."}, [[2, null, null]]]\n[2, "CALCULATE", {"expression": "__&1.input__"}, [[3, null, null]]]\n[3, "WRITE_TO_USER", {"content": "The result of the math operation is: __&2.result__"}, []]'"""

            }
        ]
        super().__init__(config, commands, command_name_to_func, recognition_messages, explanation_messages)



import re
from typing import Any, Dict

def get_indexed_data(data_name: str, generated_data_by_node: Dict[str, Any]) -> Any:
    # Get the index/key from the data name
    start_index = data_name.index('[')
    end_index = data_name.index(']')
    index = data_name[start_index + 1:end_index]
    data_name = data_name[:start_index]

    # Get the data from the generated data dictionary
    data = generated_data_by_node.get(data_name)
    if isinstance(data, (list, tuple)):
        return data[int(index)]
    elif isinstance(data, dict):
        return data[index]
    else:
        raise AssertionError(f"Could not get data for {data_name}. Indexes for type {type(data)} are not supported.")

def escape_str_for_json(value_as_str: str):
    # Replace newlines with escaped newlines
    value_as_str = value_as_str.replace(r"\n", "\\n")

    # Escape double quotes inside the string
    value_as_str = value_as_str.replace('"', '\\"')
    return value_as_str

def unescape_str_in_json(value_as_str: str):
    # Replace escaped newlines with newlines
    value_as_str = value_as_str.replace("\\n", r"\n")

    # Replace escaped double quotes with double quotes
    value_as_str = value_as_str.replace('\\"', '"')
    return value_as_str

def replace_generated_data_by_node(match_obj, generated_data_by_node: Dict[str, Any]) -> str:
    # Get the data name from the matched object
    data_name = match_obj.group(1)

    # Check if the data name has an index/key
    if '[' in data_name and ']' in data_name:
        value = get_indexed_data(data_name, generated_data_by_node)
    else:
        value = generated_data_by_node[data_name]

    value_as_str = str(value)
    value_as_str = escape_str_for_json(value_as_str)

    return value_as_str

def inject_node_data(commands_data_str: str, node_id, generated_data_by_node: dict[str, Any]):
    pattern = rf"__&{node_id}\.(\w+)__"
    return re.sub(pattern, lambda m: replace_generated_data_by_node(m, generated_data_by_node), commands_data_str)

def nullify_all_data_references(commands_data_str: str):
    """Replaces all __&i.data__ by null."""
    pattern = r"__&(\d+)\.(\w+(?:\[\d+\])*)__"
    return re.sub(pattern, "null", commands_data_str)

def find_data_references_indices(commands_data_str: str) -> dict[int, dict[str, list[tuple]]]:
    """
    Returns a dictionary with the ID of each command and the name of the generated data, along with their start and end positions
    in the raw commands data string.

    Args:
        commands_data_str (str): A string containing the data to create a graph/command.

    Returns:
        A dictionary where the keys are command IDs and the values are dictionaries representing the generated data for that
        command. These inner dictionaries have keys that are names of generated data variables and values that are lists of tuples.
        Each tuple in these lists contains the starting and ending positions of the generated data variable within the
        commands_data_str input string.

    Example:
        {1: {'urls[0]': [(115, 129), (299, 313)]}, 2: {'text': [(214, 225)]}}
    """
    pattern = r"__&(\d+)\.(\w+(?:\[\d+\])*)__"
    matches = re.finditer(pattern, commands_data_str)
    references = {}
    for match in matches:
        command_id = int(match.group(1))
        generated_data_name = match.group(2)
        start_index = match.start()
        end_index = match.end()

        if command_id not in references:
            references[command_id] = {}

        if generated_data_name not in references[command_id]:
            references[command_id][generated_data_name] = []

        references[command_id][generated_data_name].append((start_index, end_index))

    return references
